/* Copyright (c) 2004, Kentaro A. Kurahone
 * All Rights Reserved.
 */

#include "Vector.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/*
#if defined(???)
	#define HAS_SSE
#endif
*/
#if defined(__APPLE__) || defined(MACOSX)
	#include <CoreServices/CoreServices.h>
	#include <Accelerate/Accelerate.h>
	#include <vecLib/vfp.h>
	
		#define HAS_ALTIVEC
		
		/* from http://home.san.rr.com/altivec/Pages/TipArchive.html */
 		#define vec_not(A) vec_nor(A,A)
 		#define vec_nand(A,B) vec_not(vec_and(A,B))
 		#define vec_abs(A) vec_sel(A,vec_sub( vec_sub(A,A),A),vec_cmpgt(vec_sub(A,A),A))
 		#define vec_2sComp(x) vec_sub(vec_sub(x,x), x)

/*
		#include <vDsp.h>
	#if defined(__VEC__)
		#define HAS_ALTIVEC
	#endif
*/
#endif



Vector *Vector_new(void)
{
	Vector *self = calloc(1, sizeof(Vector));
	Vector_sizeTo_(self, 1);
	return self;
}

void Vector_free(Vector *self)
{
 	if (self->values) free(self->values);
 	free(self);
}

Vector *Vector_clone(Vector *self)
{
	Vector *c = Vector_new();
	Vector_copy_(c, self);
	return c;
}

Vector *Vector_subarray_(Vector *self, int localStart, int length)
{
	if (self->size <= localStart + length)
	{
		Vector *c = Vector_new();
		Vector_sizeTo_(c, length);
		memcpy(c->values, self->values, length * sizeof(NUM_TYPE));
		return c;
	}
	return 0x0;
}


void Vector_copy_(Vector *self, Vector *other)
{
	if (self->size != other->size) Vector_sizeTo_(self, other->size);
	memcpy(self->values, other->values, self->size * sizeof(NUM_TYPE));
}

int Vector_size(Vector *self)
{
	return self->size;
}

void Vector_sizeTo_(Vector *self, int size)
{
	if (self->size == size) return;
	/*
	calloc already gives 16 byte allignment
	
	#ifdef HAS_ALTIVEC
		int minSize = size > self->size ? self->size : size;
		NUM_TYPE *v = vec_malloc(newMemSize);
		memset(v, 0x0, newMemSize);
		memcpy(v, self->values, minSize);
		free(self->values);
		self->values = v;
		//self->values = vec_realloc(self->values, size * sizeof(NUM_TYPE));
	#else
	*/
		self->values = realloc(self->values, size * sizeof(NUM_TYPE));
	/*#endif*/
	self->size = size;
}

NUM_TYPE Vector_at_(Vector *self, int i)
{
	if (i < self->size) return *(self->values + i);
	return 0;
}

void Vector_at_put_(Vector *self, int i, NUM_TYPE v)
{
	if (i >= self->size) Vector_sizeTo_(self, i+1);
	*(self->values + i) = v; 
}

/* ------------------------------------------------------------------------ */

void Vector_addScalar_(Vector *self, NUM_TYPE v)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	while (n) { *d1 += v; d1++; n--; }
}

void Vector_subtractScalar_(Vector *self, NUM_TYPE v)
{
	NUM_TYPE *d1 = self->values;
	
	{
		int n = self->size;
		while (n) { *d1 -= v; d1++; n--; }
	}
}

void Vector_multiplyScalar_(Vector *self, NUM_TYPE v)
{
	NUM_TYPE *d1 = self->values;
	
	#ifdef HAS_ALTIVEC
		vsmul(d1, 1, &v, d1, 1, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n) { *d1 *= v; d1++; n--; }
		}
	#endif
}

void Vector_divideScalar_(Vector *self, NUM_TYPE v)
{
	NUM_TYPE *d1 = self->values;
	{
		int n = self->size;
		while (n) { if(v) *d1 /= v; else *d1 = 0; d1++; n--; }
	}
}

/* ------------------------------------------------------------------------ */

NUM_TYPE Vector_sum(Vector *self)
{
	NUM_TYPE total = 0;
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	
	while (n)
	{
		total += *d1;
		d1++; 
		n --;
	}
	return total;
}

int Vector_addArray_(Vector *self, Vector *other)
{
	NUM_TYPE *d1 = self->values;
	NUM_TYPE *d2 = other->values;
	
	if (self->size != other->size) return -1;

	#ifdef HAS_ALTIVEC
		vadd(d1, 1, d2, 1, d1, 1, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n)
		{
			*d1 += *d2;
			d1++; d2++;
			n --;
		}
		}
	#endif
	return 0;
}
	
int Vector_subtractArray_(Vector *self, Vector *other)
{
	NUM_TYPE *d1 = self->values;
	NUM_TYPE *d2 = other->values;
	
	if (self->size != other->size) return -1;
	
	#ifdef HAS_ALTIVEC
		vsub(d1, 1, d2, 1, d1, 1, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n)
		{
			*d1 -= *d2;
			d1++; d2++;
			n --;
		}
		}
	#endif
	return 0;
}
 
int Vector_multiplyArray_(Vector *self, Vector *other)
{
	NUM_TYPE *d1 = self->values;
	NUM_TYPE *d2 = other->values;
	
	if (self->size != other->size) return -1;
	
	#ifdef HAS_ALTIVEC
		vmul(d1, 1, d2, 1, d1, 1, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n)
		{
			*d1 *= *d2;
			d1++; d2++;
			n --;
		}
		}
	#endif
	return 0;
}

int Vector_divideArray_(Vector *self, Vector *other)
{
	NUM_TYPE *d1 = self->values;
	NUM_TYPE *d2 = other->values;
	
	if (self->size != other->size) return -1;
	{
		int n = self->size;
		while (n)
		{
			if(*d2) *d1 /= *d2;
			else *d1 = 0;
			d1++; d2++;
			n --;
		}
	}
	return 0;
}

NUM_TYPE Vector_dotProduct_(Vector *self, Vector *other)
{
	NUM_TYPE *d1 = self->values;
	NUM_TYPE *d2 = other->values;
	NUM_TYPE result = 0;
	
	if (self->size != other->size) 
	{
		printf("Error: Vector_dotProduct_() sizes not equal\n");
		return -1; 
	}
	
	#ifdef HAS_ALTIVEC
		dotpr(d1, 1, d2, 1, &result, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n)
		{
			result += (*d1) * (*d2);
			d1++; d2++;
			n --;
		}
		}
	#endif
	return result;
}

void Vector_square(Vector *self)
{
	NUM_TYPE *d1 = self->values;
	
	#ifdef HAS_ALTIVEC
		vsq(d1, 1, d1, 1, self->size);
	#elif HAS_SSE
	#else
		{
		int n = self->size;
		while (n)
		{
			*d1 *= (*d1);
			d1++; 
			n --;
		}
		}
	#endif
}

void Vector_normalize(Vector *self)
{
	NUM_TYPE v1 = Vector_max(self);
	NUM_TYPE v2 = (NUM_TYPE)fabs((double)Vector_min(self));
	NUM_TYPE v = v1 > v2 ? v1 : v2;
	
	if (!v) return;
	Vector_multiplyScalar_(self, 1.0/v);
}

void Vector_absolute(Vector *self)
{
	NUM_TYPE *d1 = self->values;

/*
	#ifdef HAS_ALTIVEC
		vfabf(d1);
	#else
	*/
			{	
			int n = self->size;
			while (n)
			{
	if (*d1 < 0) *d1 = -(*d1);
	d1++; 
	n --;
			}
			}
}

void Vector_log(Vector *self)
{
	NUM_TYPE *d1 = self->values;

	/*
	#ifdef HAS_ALTIVEC
		vlogf(d1);
	#else	
	*/
			{	
			int n = self->size;
			while (n)
			{
	*d1 = (NUM_TYPE)log((double)(*d1));
	d1++; 
	n --;
			}
			}
}

void Vector_log10(Vector *self)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;

	
	while (n)
	{
		*d1 = (NUM_TYPE)log10((double)(*d1));
		d1++; 
		n --;
	}
}

void Vector_pow(Vector *self, int p)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;

	/*
	#ifdef HAS_ALTIVEC
		vipowf(d1, p);
	#else
	*/
	while (n)
	{
		*d1 = (NUM_TYPE)pow((double)(*d1), p);
		d1++; 
		n --;
	}
}

void Vector_random(Vector *self, float min, float max)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	float range = max - min;

	
	while (n)
	{
		*d1 = min + range * ((NUM_TYPE)rand())/((NUM_TYPE)RAND_MAX);
		d1++; 
		n --;
	}
}

int Vector_ElementCmp(const void *a, const void *b)
{ return *((NUM_TYPE *)a) < *((NUM_TYPE *)b); }

void Vector_sort(Vector *self)
{
	if (!self->values) return;
	qsort(self->values, self->size, sizeof(NUM_TYPE), Vector_ElementCmp);
}

int Vector_ElementReverseCmp(const void *a, const void *b)
{ return *((NUM_TYPE *)a) > *((NUM_TYPE *)b); }

void Vector_reverseSort(Vector *self)
{
	if (!self->values) return;
	qsort(self->values, self->size, sizeof(NUM_TYPE), Vector_ElementReverseCmp);
}

/* ---------------------------------------------------------- */

NUM_TYPE Vector_min(Vector *self)
{
	NUM_TYPE m;
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	if (!d1) return 0;
	
	m = *d1;
	while (n)
	{
		if (*d1 < m) m = *d1;
		d1++; 
		n --;
	}
	return m;
}

NUM_TYPE Vector_max(Vector *self)
{
	NUM_TYPE m;
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	if (!d1) return 0;
	
	m = *d1;
	while (n)
	{
		if (*d1 > m) m = *d1;
		d1++; 
		n --;
	}
	return m;
}

NUM_TYPE Vector_mean(Vector *self)
{
	NUM_TYPE total = 0;
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	if (!d1) return 0;
	
	while (n)
	{
		total += (*d1);
		d1++; 
		n --;
	}
	return total / ((NUM_TYPE)self->size);
}

NUM_TYPE Vector_meanSquare(Vector *self)
{
	NUM_TYPE total = 0;
	int n = self->size;
	NUM_TYPE *d1 = self->values;
	if (!d1) return 0;
	
	while (n)
	{
		total += (*d1)*(*d1);
		d1++; 
		n --;
	}
	return total / ((NUM_TYPE)self->size);
}

void Vector_setMin_(Vector *self, NUM_TYPE v)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;

	
	while (n)
	{
		if (*d1 < v) *d1 = v;
		d1++; 
		n --;
	}
}

void Vector_setMax_(Vector *self, NUM_TYPE v)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;

	
	while (n)
	{
		if (*d1 > v) *d1 = v;
		d1++; 
		n --;
	}
}

void Vector_set_(Vector *self, NUM_TYPE v)
{
	int n = self->size;
	NUM_TYPE *d1 = self->values;

	
	while (n)
	{
		*d1 = v;
		d1++; 
		n --;
	}
}

NUM_TYPE Vector_rootMeanSquare(Vector *self)
{ return sqrt(Vector_meanSquare(self)); }

#define PI 3.14159265

void Vector_sin(Vector *self, 
	int start, int length, 
	NUM_TYPE amplitude, NUM_TYPE offset, NUM_TYPE wavelength)
{
	NUM_TYPE frequency = 2.0 * PI / wavelength ;
	int size = self->size;
	int max = start + length;	
	NUM_TYPE *d1 = self->values;
	int n = start;
	
	if (n < 0) n = 0;
	if (n > size) return;

	if (max <= n) return;
	if (max > size) max = size;
	
	while (n < max)
	{ 
			d1[n] += amplitude * (float)sin((double)(offset + (n - start) * frequency)); 
			n++;
	}
}

void Vector_vsinf(Vector *self)
{
	NUM_TYPE *d1 = self->values;

	/*
	#ifdef HAS_ALTIVEC
		vsinf(d1);
	#else
	*/
	{
	int n = self->size;
	while (n)
	{
		*d1 = (NUM_TYPE)sin((double)*d1);
		d1++; 
		n --;
	}
	}
}

void Vector_vcosf(Vector *self)
{
	NUM_TYPE *d1 = self->values;

	/*
	#ifdef HAS_ALTIVEC
		vcosf(d1);
	#else
	*/
	{
	int n = self->size;
	while (n)
	{
		*d1 = (NUM_TYPE)cos((double)*d1);
		d1++; 
		n --;
	}
	}
}

void Vector_vtanf(Vector *self)
{
	NUM_TYPE *d1 = self->values;

	/*
	#ifdef HAS_ALTIVEC
		vtanf(d1);
	#else
	*/
	{
	int n = self->size;
	while (n)
	{
		*d1 = (NUM_TYPE)tan((double)*d1);
		d1++; 
		n --;
	}
	}
}
