/* Copyright (c) 2004, Steve Dekorte
 * All Rights Reserved.
 */

#include "IoState.h"
#include "IoObject.h"
#include "IoNil.h"
#include "IoVector.h"
#include "IoNumber.h"
#include "base/Vector.h"

#define VIVAR(self) ((Vector *)(self->data))

void *IoMessage_locals_vectorArgAt_(IoMessage *self, void *locals, int n)
{
  IoObject *v = IoMessage_locals_valueArgAt_(self, locals, n);
  if (!ISVECTOR(v)) 
  { IoMessage_locals_numberArgAt_errorForType_(self, locals, n, "Vector"); }
  return v;
}

IoTag *IoVector_tag(void *state)
{
  IoTag *tag = IoTag_newWithName_("Vector");
  tag->state = state;
  tag->cloneFunc = (TagCloneFunc *) IoVector_rawClone;
  tag->markFunc  = (TagMarkFunc *)  IoVector_mark;
  tag->freeFunc  = (TagFreeFunc *)  IoVector_free;
  return tag;
}

IoVector* IoVector_proto(void *state)
{
  IoMethodTable methodTable[] = {
    {"copy",  IoVector_copy_},
    {"subarray",  IoVector_subarray},
    {"size",      IoVector_size},
    {"setSize",   IoVector_setSize_},
    {"at",        IoVector_at_},
    {"atPut",     IoVector_at_put_},
    
    {"+=",        IoVector_add},
    {"-=",        IoVector_subtract},
    {"*=",        IoVector_multiply},
    {".",         IoVector_dotProduct},
    
    {"sum",       IoVector_sum},
    {"min",       IoVector_min},
    {"max",       IoVector_max},
    {"mean",      IoVector_mean},
    {"square",    IoVector_square},
    {"meanSquare",       IoVector_meanSquare},
    {"rootMeanSquare",   IoVector_rootMeanSquare},
    {"normalize",        IoVector_normalize},
    {"absolute",         IoVector_absolute},

    {"log",       IoVector_log},
    {"log10",     IoVector_log10},
    {"pow",       IoVector_pow},
    {"random",    IoVector_random},
    {"sort",      IoVector_sort},
    {"reverseSort", IoVector_reverseSort},
    
    
    {"setMin",    IoVector_setMin_},
    {"setMax",    IoVector_setMax_},
    {"set",    IoVector_set_},
   
    {"sin",       IoVector_sin},
    {NULL, NULL}
  };

  IoObject *self = IoObject_new(state);
  self->tag = IoVector_tag(state);
  self->data = Vector_new();

  IoState_registerProtoWithFunc_(state, self, IoVector_proto);
  IoObject_addMethodTable_(self, methodTable);

  return self;
}

IoVector* IoVector_rawClone(IoVector *self)
{
  IoObject *child = IoObject_rawClonePrimitive(self);
  child->data = Vector_clone(VIVAR(self));
  return child;
}

IoVector *IoVector_new(void *state)
{
  IoObject *proto = IoState_protoWithInitFunction_(state, IoVector_proto);
  return IoObject_rawClone(proto);
}

void IoVector_setVector_(IoVector *self, Vector *na)
{
  if (VIVAR(self) != na) Vector_free(VIVAR(self));
  VIVAR(self) = na;
}

Vector *IoVector_rawVector(IoVector *self) { return VIVAR(self); }


void IoVector_free(IoVector *self)
{
  Vector_free(VIVAR(self));
}

void IoVector_mark(IoVector *self)
{
}

IoObject *IoVector_size(IoVector *self, IoObject *locals, IoMessage *m)
{
  return IONUMBER(Vector_size(VIVAR(self)));
}

IoObject *IoVector_setSize_(IoVector *self, IoObject *locals, IoMessage *m)
{
  int n = IoMessage_locals_intArgAt_(m, locals, 0);
  Vector_sizeTo_(VIVAR(self), n);
  return self;
}

IoObject *IoVector_copy_(IoVector *self, IoObject *locals, IoMessage *m)
{
  IoVector *other = IoMessage_locals_vectorArgAt_(m, locals, 0);
  Vector_copy_(VIVAR(self), VIVAR(other));
  return self;
}

IoObject *IoVector_subarray(IoVector *self, IoObject *locals, IoMessage *m)
{
  int start = IoMessage_locals_intArgAt_(m, locals, 0);
  int length = IoMessage_locals_intArgAt_(m, locals, 1);
  Vector *a = Vector_subarray_(VIVAR(self), start, length);
  if (a)
  {
    IoVector *ioa = IoVector_new(IOSTATE);
    IoVector_setVector_(ioa, a);
    return ioa;
  }
  return IONIL(self);
}

IoObject *IoVector_at_(IoVector *self, IoObject *locals, IoMessage *m)
{
  int n = IoMessage_locals_intArgAt_(m, locals, 0);
  double d = Vector_at_(VIVAR(self), n);
  return IONUMBER(d);
}

IoObject *IoVector_at_put_(IoVector *self, IoObject *locals, IoMessage *m)
{
  int n = IoMessage_locals_intArgAt_(m, locals, 0);
  double d = IoMessage_locals_doubleArgAt_(m, locals, 1);
  Vector_at_put_(VIVAR(self), n, d);
  return self;
}

/* --- math -------------------------------------------- */

IoObject *IoVector_add(IoVector *self, IoObject *locals, IoMessage *m)
{
  IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
  
  if (ISVECTOR(other))
  { Vector_addArray_(VIVAR(self), VIVAR(other)); }
  else if (ISNUMBER(other))
  { Vector_addScalar_(VIVAR(self), CNUMBER(other)); }
  else
  {
    IoState_error_description_(IOSTATE, m, 
      "IoVector add", "requires Vector or Number argument"); 
  }
  return self;
}

IoObject *IoVector_subtract(IoVector *self, IoObject *locals, IoMessage *m)
{
  IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
  
  if (ISVECTOR(other))
  { Vector_subtractArray_(VIVAR(self), VIVAR(other)); }
  else if (ISNUMBER(other))
  { Vector_subtractScalar_(VIVAR(self), CNUMBER(other)); }
  else
  {
    IoState_error_description_(IOSTATE, m, 
      "IoVector subtract", "requires Vector or Number argument"); 
  }
  return self;
}

IoObject *IoVector_multiply(IoVector *self, IoObject *locals, IoMessage *m)
{
  IoObject *other = IoMessage_locals_valueArgAt_(m, locals, 0);
  
  if (ISVECTOR(other))
  { Vector_multiplyArray_(VIVAR(self), VIVAR(other)); }
  else if (ISNUMBER(other))
  { Vector_multiplyScalar_(VIVAR(self), CNUMBER(other)); }
  else
  {
    IoState_error_description_(IOSTATE, m, 
      "IoVector multiply", "requires Vector or Number argument"); 
  }
  return self;
}

IoObject *IoVector_dotProduct(IoVector *self, IoObject *locals, IoMessage *m)
{
  IoObject *other = IoMessage_locals_vectorArgAt_(m, locals, 0);
  double r = Vector_dotProduct_(VIVAR(self), VIVAR(other));
  return IONUMBER(r);
}

IoObject *IoVector_sum(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_sum(VIVAR(self))); }

IoObject *IoVector_min(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_min(VIVAR(self))); }

IoObject *IoVector_max(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_max(VIVAR(self))); }

IoObject *IoVector_mean(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_mean(VIVAR(self))); }


IoObject *IoVector_square(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_square(VIVAR(self)); return self; }

IoObject *IoVector_normalize(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_normalize(VIVAR(self)); return self; }

IoObject *IoVector_absolute(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_absolute(VIVAR(self)); return self; }


IoObject *IoVector_log(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_log(VIVAR(self)); return self; }

IoObject *IoVector_log10(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_log10(VIVAR(self)); return self; }

IoObject *IoVector_pow(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  int p = IoMessage_locals_intArgAt_(m, locals, 0);
  Vector_pow(VIVAR(self), p); 
  return self; 
}

IoObject *IoVector_random(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  int min = IoMessage_locals_intArgAt_(m, locals, 0);
  int max = IoMessage_locals_intArgAt_(m, locals, 1);
  Vector_random(VIVAR(self), min, max); 
  return self; 
}

IoObject *IoVector_sort(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_sort(VIVAR(self)); return self; }

IoObject *IoVector_reverseSort(IoVector *self, IoObject *locals, IoMessage *m)
{ Vector_reverseSort(VIVAR(self)); return self; }




IoObject *IoVector_meanSquare(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_meanSquare(VIVAR(self))); }

IoObject *IoVector_rootMeanSquare(IoVector *self, IoObject *locals, IoMessage *m)
{ return IONUMBER(Vector_rootMeanSquare(VIVAR(self))); }

IoObject *IoVector_setMin_(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  double v = IoMessage_locals_doubleArgAt_(m, locals, 0);
  Vector_setMin_(VIVAR(self), v);
  return self; 
}

IoObject *IoVector_setMax_(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  double v = IoMessage_locals_doubleArgAt_(m, locals, 0);
  Vector_setMax_(VIVAR(self), v);
  return self; 
}

IoObject *IoVector_set_(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  double v = IoMessage_locals_doubleArgAt_(m, locals, 0);
  Vector_set_(VIVAR(self), v);
  return self; 
}

IoObject *IoVector_sin(IoVector *self, IoObject *locals, IoMessage *m)
{ 
  int start = IoMessage_locals_intArgAt_(m, locals, 0);
  int length = IoMessage_locals_intArgAt_(m, locals, 1);
  double amplitude = IoMessage_locals_doubleArgAt_(m, locals, 2);
  double offset = IoMessage_locals_doubleArgAt_(m, locals, 3);
  double wavelength = IoMessage_locals_doubleArgAt_(m, locals, 4);
  
  Vector_sin(VIVAR(self), start, length, amplitude, offset, wavelength);
  return self;
}
