/*
 EventManager ioDoc(
				docCopyright("Steve Dekorte", 2002)
				docLicense("BSD revised")
				docDescription("A binding for libevent.")
				docCategory("Networking")
  */

#include "IoEventManager.h"
#include "IoNumber.h"
#include "Common.h"
#include <signal.h>
#include <math.h>
#if defined(__APPLE__) && defined(__MACH__)
#include <AvailabilityMacros.h>
#endif

#include <time.h>
#ifndef WIN32
#include <unistd.h>
#endif

#define DATA(self) ((IoEventManagerData *)IoObject_dataPointer(self))

IoTag *IoEventManager_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("EventManager");
	tag->state = state;
	tag->cloneFunc = (TagCloneFunc *)IoEventManager_rawClone; 
	tag->freeFunc  = (TagFreeFunc *)IoEventManager_free;
	tag->markFunc  = (TagMarkFunc *)IoEventManager_mark;
	return tag;
}

#include <stdlib.h>

IoEventManager *IoEventManager_proto(void *state)
{
	IoObject *self = IoObject_new(state);
	self->tag = IoEventManager_tag(state);
	
	IoObject_setDataPointer_(self, calloc(1, sizeof(IoEventManagerData)));    
	
	DATA(self)->handleEventMessage = IoMessage_newWithName_(state, IOSYMBOL("handleEvent"));
	
	IoState_registerProtoWithFunc_((IoState *)state, self, IoEventManager_proto);
	
	{
		IoMethodTable methodTable[] = {
		{"addEvent", IoEventManager_addEvent},
		{"removeEvent", IoEventManager_removeEvent},
			
		{"listen", IoEventManager_listen},
		{"listenUntilEvent", IoEventManager_listenUntilEvent},
		{"setListenTimeout", IoEventManager_setListenTimeout},
		
		{NULL, NULL},
		};
		
		IoObject_addMethodTable_(self, methodTable);
	}
	

//#if !defined(AVAILABLE_MAC_OS_X_VERSION_10_4_AND_LATER)
#if defined(__APPLE__)
	setenv("EVENT_NOKQUEUE", "1", 1);
	//printf("EventManager warning: disabling libevent kqueue support to avoid bug in OSX < 10.4\n"); 
	setenv("EVENT_NOPOLL", "1", 1);
	//setenv("EVENT_NOSELECT", "1", 1);
#endif

	DATA(self)->eventBase = event_init();
	
	return self;
}

// ----------------------------------------------------------- 

IoEventManager *IoEventManager_rawClone(IoEventManager *self)
{ 
	return self; 
}

void IoEventManager_mark(IoEventManager *self) 
{
	IoObject_shouldMark(DATA(self)->handleEventMessage);
	
	// add code to walk event list and mark context values
}

void IoEventManager_free(IoEventManager *self) 
{ 
	// we don't free libevent since it's global and there
	// may be other stuff (possibly other IoStates) in the process
	// using it
	
	free(IoObject_dataPointer(self)); 
}

void IoEventManager_BrokenPipeSignalHandler(int v)
{
	//printf("IoEventManager catching broken pipe signal\n");
}

IoEventManager *IoEventManager_new(void)
{
	IoEventManager *self = (IoEventManager *)calloc(1, sizeof(IoEventManager));
	
	event_init();
	
#ifndef WIN32
     // Some unixes raise a signal when a socket pipe is broken, 
	// so we do this to avoid a crash.
	signal(SIGPIPE, IoEventManager_BrokenPipeSignalHandler);
#endif
	
    //IoEventManager_setDescriptorLimitToMax(self);
	Socket_SetDescriptorLimitToMax();
	
	return self;
}

/*
static int validEventType(int t)
{	
	return (t == EV_READ || t == EV_WRITE || t == EV_TIMEOUT || t == EV_SIGNAL);
}
*/

void IoEvent_handleEvent(int fd, short eventType, void *context)
{
	IoEvent *self = (IoEvent *)context;

	event_del(IoEvent_rawEvent(self)); 
	
	if (!RawDescriptor_isValid(fd))
	{
		printf("IoEvent_handleEvent: handleEvent on bad file descriptor\n");
	}
	
	IoState_pushRetainPool(IOSTATE);
	{
	IoEventManager *em = IoState_protoWithInitFunction_(IOSTATE, IoEventManager_proto);
	IoMessage *m = DATA(em)->handleEventMessage;
	
	IoMessage_setCachedArg_to_(m, 0, IOBOOL(self, eventType == EV_TIMEOUT));
	IoMessage_locals_performOn_(m, self, self);
	}
	IoState_popRetainPool(IOSTATE);
}

static struct timeval timevalFromDouble(double seconds)
{
	double ip;
	double fraction = modf(seconds, &ip);
	struct timeval tv;
	tv.tv_sec = ip;
	tv.tv_usec = fraction/1000000.0;
	return tv; 
}

IoObject *IoEventManager_addEvent(IoEventManager *self, IoObject *locals, IoMessage *m)
{ 	
	IoEvent *event = IoMessage_locals_eventArgAt_(m, locals, 0);
	struct event *ev = IoEvent_rawEvent(event);

	int fd = IoMessage_locals_intArgAt_(m, locals, 1);
	
	int eventType = IoMessage_locals_intArgAt_(m, locals, 2);
	
	double timeout = IoMessage_locals_doubleArgAt_(m, locals, 3);
	struct timeval tv = timevalFromDouble(timeout);

	if (!RawDescriptor_isValid(fd))
	{
		IoState_error_(IOSTATE, m, "IoEventManager_addEvent: attempt to add bad file descriptor %i", fd); 
	}
	
	event_set(ev, fd, eventType, IoEvent_handleEvent, event);
	event_base_set(DATA(self)->eventBase, ev);
	
	if (timeout < 0)
	{
		event_add(ev, NULL); // no timeout
	}
	else
	{
		event_add(ev, &tv);
	}
	
	return self;
}

IoObject *IoEventManager_removeEvent(IoEventManager *self, IoObject *locals, IoMessage *m)
{ 	
	IoEvent *event = IoMessage_locals_eventArgAt_(m, locals, 0);
	event_del(IoEvent_rawEvent(event)); 
	return self;
}

// ------------------------------------

IoObject *IoEventManager_setListenTimeout(IoEventManager *self, IoObject *locals, IoMessage *m)
{ 	
	double timeout = IoMessage_locals_doubleArgAt_(m, locals, 0);
	struct timeval tv = timevalFromDouble(timeout);
	event_loopexit(&tv);
	return self;
}

IoObject *IoEventManager_listen(IoEventManager *self, IoObject *locals, IoMessage *m)
{
	int hadEvents = event_base_loop(DATA(self)->eventBase, EVLOOP_NONBLOCK);
	return IONUMBER(hadEvents);
}

IoObject *IoEventManager_listenUntilEvent(IoEventManager *self, IoObject *locals, IoMessage *m)
{
	int hadEvents = event_base_loop(DATA(self)->eventBase, EVLOOP_ONCE);
	
	if (hadEvents == -1)
	{
		//printf("hadEvents == -1\n");
		IoState_error_(IOSTATE, m, "EventManager: error in event_base_loop"); 
	}
	
	return IONUMBER(hadEvents);
}
