
Sequence parseAsUrlToMap := method(
    m := Map clone
    s := self
    protocol := "http"
    
	if(s containsSeq("://"),
    	protocol := s beforeSeq("://")
    	s := s afterSeq("://")
    )
    if(s == nil, return)
    host := s beforeSeq("/")
    if(host containsSeq(":")) then(
        port := host afterSeq(":")
        host := host  beforeSeq(":")
    ) else(
        port := "80"
    )

    request := s afterSeq("/")

    m atPut("protocol", protocol)
    m atPut("host", host)
    m atPut("port", port)
    m atPut("request", request)
    m atPut("query", if(request, request afterSeq("?"), nil))
    
    m
)

URL := Object clone do(
	type := "URL"
	docCategory("Networking")
	docCopyright("Steve Dekorte", 2004)
	docLicense("BSD revised")
	docDescription("The URL object is usefull for fetching web pages and parsing URLs. Example;
	<pre>
	page := URL clone setURL(\"http://www.google.com/\") fetch
	</pre>")

    docSlot("url", "Returns url string.")
    url := ""

    docSlot("setURL(urlString)",
    "Sets the url string and parses into the protocol, host, port path, and query slots. Returns self.")

	newSlot("socketProto", Socket clone)
	
	init := method(
		self socket := socketProto clone
	)
	
	with := method(s, self clone setURL(s))
	
    setURL := method(s, 
    	if(s lastPathComponent pathExtension == "" and s endsWithSeq("/") == false, s = s .. "/")
        url = s asString
        parse
        self
    )
    
    docSlot("error", "Returns error string if it has been set or nil otherwise.")
    newSlot("error", nil)
    newSlot("host", nil)
    newSlot("protocol", nil)
    
    unparse := method(
		if(host) then(
    		d := if(protocol, protocol .. "://", "") .. host .. if(port and port != 80, ":" .. port, "")
    	) else(
    		d := ""
    	)
    	self url := d .. path .. if(query, "?" .. query, "")
    	url
    )
    
    clear := method(
        self do(
            protocol := nil
            host  := nil
            port  := 80
            path  := "/"
            query := nil
            error := nil
        )
    )

    docSlot("escapeString(aString)", "Returns a new String that is aString with the appropriate characters replaced by their URL escape codes.")

    escapeString := method(u,
        u := u clone asMutable 
        u replaceSeq("%","%25")
        escapeCodes foreach(k, v, u replaceSeq(k, v))
        u asString
    )

    docSlot("unescapeString(aString)", "Returns a new String that is aString with the URL escape codes replaced by the appropriate characters.")
    
    unescapeString := method(u,
        u := u clone asMutable 
        escapeCodes foreach(k, v, u replaceSeq(v, k))
        u replaceSeq("%25", "%")
        u asString
    )

    docSlot("escapeCodes", "Returns a Map whose key/value pairs are special characters and their URL escape codes.")
    
    escapeCodes := Map clone do(
        atPut(" ","%20")
        atPut("<","%3C")
        atPut(">","%3E")
        atPut("#","%23")
        //atPut("%","%25")
        atPut("{","%7B")
        atPut("}","%7D")
        atPut("|","%7C")
        //atPut("backslash","%5C")
        atPut("^","%5E")
        atPut("~","%7E")
        atPut("[","%5B")
        atPut("]","%5D")
        atPut("\`","%60")
        atPut(";","%3B")
        atPut("/","%2F")
        atPut("?","%3F")
        atPut(":","%3A")
        atPut("@","%40")
        atPut("=","%3D")
        atPut("&","%26")
        atPut("$","%24")
    )

    docSlot("referer", "Returns the referer String or nil if not set.")
    docSlot("setReferer(aString)", "Sets the referer. Returns self.")
    newSlot("referer", nil)

    parse := method(
        clear
        m := url parseAsUrlToMap 
        
        if(m at("request")) then(
          m atPut("request", "/" .. m at("request"))
        ) else(
          m atPut("request", "/")
        )
        m atPut("path", m at("request"))
        m foreach(k, v, self setSlot(k, v); /*write(k, "=", v, "\n")*/)
        if(port) then(port = port asNumber) else(port = 80)
    )

    docSlot("fetch", "Fetches the url and returns the result as a Sequence. Raises an exception if an error occurs.")
    
    fetch := method(url,
        if(url, setURL(url))
        if(protocol == "http", return fetchHttp)
        Exception raise("protocol '" .. protocol .. "' unsupported")
    )

    docSlot("fetchWithProgress(progressBlock)", "Same as fetch, but with each read, progressBlock is called with the readBuffer and the content size as parameters.")

    fetchWithProgress := method(progressBlock,
        if(protocol == "http", return fetchHttpWithProgress(getSlot("progressBlock")))
        Exception raise("protocol '" .. protocol .. "' unsupported")
    )

    docSlot("stopFetch", "Stops the fetch, if there is one. Returns self.")

    stopFetch := method(socket close; self)

    docSlot("requestHeader", "Returns a Sequence containing the request header that will be sent.")

    requestHeader := method(
        header := Sequence clone
        //write("request = [", request, "]\n")
        header appendSeq("GET ", request," HTTP/1.0\r\n")
        header appendSeq("User-Agent: Mozilla/5.0\r\n")
        header appendSeq("Host: ", host, "\r\n")
        if(referer, header appendSeq("Referer: ", referer, "\r\n"))
        header appendSeq("Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, image/tiff, multipart/x-mixed-replace, */*\r\n")
        header appendSeq("Accept-Charset: iso-8859-1, utf-8, iso-10646-ucs-2, macintosh, windows-1252, *\r\n")
        header appendSeq("Accept-Language: en, *;q=0.5\r\n")
        header appendSeq("\r\n")
        header
    )

    fetchHttp := method(
        socket setHost(host) setPort(port) connect 
        if(socket error, error = socket error; return nil) 
    
        socket streamWrite(requestHeader)
        while(socket streamReadNextChunk, nil)
        socket close
        if(socket error, writeln(socket error); return) 
        //write("read ", socket readBuffer size, " bytes\n")
    
        b := socket readBuffer
        b clipBeforeEndOfSeq("\r\n\r\n") ifNil(b clipBeforeEndOfSeq("\n\n"))
        b
    )

    fetchHttpWithProgress := method(progressBlock,
		socket setHost(host) setPort(port) connect
        if(socket error, error := socket error; return nil)
    
        socket streamWrite(requestHeader)
    
        socket streamReadNextChunk
        b := socket readBuffer
        self contentLength := 0
        beg := b findSeq( "Content-Length:" )
        if(beg,
            end := b findSeq( "\r", beg + 16 )
            self contentLength = b slice( beg + 16, end ) asNumber
        )
    
        while(socket streamReadNextChunk, progressBlock(b size, contentLength))
        if(socket error, write(socket error, "\n"); return)
        //write("read ", socket readBuffer size, " bytes\n")
        socket close
    
        //write("buffer[", b, "]\n")
        b clipBeforeEndOfSeq("\r\n\r\n") ifNil(b clipBeforeEndOfSeq("\n\n"))
    
        return b
    )

	docSlot("fetchToFile(aFile)", "fetch the url and save the result to the specified File object. Saving is done as the data is read, which help minimize memory usage. Returns self on success or nil on error.")

    fetchToFile := method(file,
        ip := Host clone setName(host) address
        socket setHost(ip) setPort(port) 
        socket connect 
    
        if(socket error, error := socket error; return nil) 
    
        socket streamWrite(requestHeader)
        b := socket readBuffer
        
        while(socket readNextChunk, 
            if(b containsSeq("\r\n\r\n") or b containsSeq("\n\n"), break)
        )
        
        if(socket error) then(write(socket error, "\n"); return )
    
        if(b clipBeforeEndOfSeq("\r\n\r\n") == nil, b clipBeforeEndOfSeq("\n\n"))
        
        file write(b)
        b empty
    
        while(socket readNextChunk, 
            file write(b)
            b empty
        )
        socket close
        
        if(socket error, self setError(socket error); return nil)     
        self
    )
    
    
    childUrl := method(u, 
        if(u beginsWithSeq("http") not, 
            u = Path with(url pathComponent, u)
        )
        URL clone setURL(u) setReferer(url)
    )

    docSlot("post(data)", "Sends an http post message. If data is a Map, it's key/value pairs are send as the post parameters. If data is a Sequence or String, it is sent directly. Returns self on success or nil on error.")

    post := method(postdata,
        ip := Host clone setName(host) address
        
        header := Sequence clone
        header appendSeq(
            "POST ", request, " HTTP/1.0\r\n",
            "User-Agent: Mozilla/4.0\r\n",
            "Host: ", host, "\r\n",
            "Accept: */*\r\n",
            "Content-Type: application/x-www-form-urlencoded\r\n"
        )
        
        if(postdata type == "Map") then(
            buffer := Sequence clone
            postdata keys foreach(i, j,
            buffer appendSeq(escapeString(j), "=", escapeString(postdata at(j)))
            if(i < postdata size - 1, buffer appendSeq("&"))
            )
            content := buffer asString
        ) else(
            content := postdata
        )
        
        header appendSeq("Content-Length: ", content size, "\r\n\r\n", content)
        
        socket setHost(ip) setPort(port) 
        socket connect 
        
        if(socket error, 
            error = socket error
            return nil
        )
         
        socket streamWrite(header)
        while (socket readNextChunk, socket readBuffer empty)
        socket close
        if(socket error, self setError(socket error); return nil)     
        self
    )

    openOnDesktop := method(
        platform := System platform
        if(platform == "Mac OS/X" ) then(
            System system( "open " .. url)
        ) elseif(platform == "Windows NT" ) then(
            System system( "cmd /c start " .. url)
        ) elseif(platform == "Windows 9X" ) then(
            System system( "command /c start " .. url)
        ) else(
            // assume generic Unix?
            System system( "open " .. url)
        )
    )
)

Object doURL := method(url, self doString(URL clone setURL(url) fetch))
