/*   
    copyright: Steve Dekorte, 2004. All rights reserved.
    license: See _BSDLicense.txt.
*/
 
#include "SearchEngine.h"

SearchEngine *SearchEngine_new(void)
{
    SearchEngine *self = (SearchEngine *)calloc(1, sizeof(SearchEngine));
    /*SearchEngine_setPath_(self, "default");*/
    return self;
}

void SearchEngine_free(SearchEngine *self)
{
    SearchEngine_close(self);
    if (self->path) free(self->path);
    if (self->logPath) free(self->logPath);
    if (self->buf) free(self->buf);
    free(self);
}

void SearchEngine_setPath_(SearchEngine *self, char *path)
{
    self->path    = strcpy((char *)realloc(self->path,    strlen(path)+1), path);
    self->logPath = strcpy((char *)realloc(self->logPath, strlen(path)+5), path);
    strcat(self->logPath,  ".log");
}

char *SearchEngine_fileName(SearchEngine *self)
{
#ifdef __WIN32__
    char *fileName = strrchr(self->path, '\\');
#else
    char *fileName = strrchr(self->path, '/');
#endif
    return  fileName ? fileName : self->path;
}

void SearchEngine_setLogPath_(SearchEngine *self, char *path)
{
    char *fileName = SearchEngine_fileName(self);
    self->logPath = strcpy((char *)realloc(self->logPath, strlen(path) + strlen(fileName)+10), path);
    if (self->logPath[strlen(self->logPath)-1] != '/') strcat(self->logPath, "/");
    strcat(self->logPath, fileName);
    strcat(self->logPath,  ".log");
}

void SearchEngine_setPath_withExtension_(SearchEngine *self, char *path, char *ext)
{
    size_t length = strlen(path) + 1 + strlen(ext) + 1;
    char *s = (char *)calloc(1, length);
    strcat(s, path);
    strcat(s, ".");
    strcat(s, ext);
    SearchEngine_setPath_(self, s);
    free(s);
}

char *SearchEngine_path(SearchEngine *self)
{
    return self->path;
}

void SearchEngine_open(SearchEngine *self)
{
    self->file = fopen(self->path, "r+b");
    
    if (!self->file) 
    {
	self->file = fopen(self->path, "wb");
	fclose(self->file);
	self->file = fopen(self->path, "r+b");
    }

    fseek(self->file, 0, SEEK_END);
    self->maxPos = ftell(self->file);

    self->log = fopen(self->logPath, "r+b");
    /*printf("%s\n", self->logPath);*/
    
    if (!self->log) 
    {
	self->log = fopen(self->logPath, "wb");
	fclose(self->log);
	self->log = fopen(self->logPath, "r+b");
    }
        
    self->header.state = SearchEngine_UNCOMMITTED;
    self->header.writeCount = 0;
    SearchEngine_readHeader(self);
    
    /*
    if (SearchEngine_isCommitted(self)) { SearchEngine_sync(self); }
    */

    self->logPos = sizeof(SearchEngineLogHeader);
}

void SearchEngine_close(SearchEngine *self)
{
    if (self->file) 
    {
	fclose(self->file);
	self->file = 0x0;
    }
    
    if (self->log) 
    {
	fclose(self->log);
	self->log = 0x0;
    }
}

void SearchEngine_delete(SearchEngine *self)
{
    remove(self->path);
}

size_t SearchEngine_fwrite(SearchEngine *self, void *buf, size_t size, size_t nobjs)
{
    size_t result;
    size_t total = size * nobjs;

    #ifdef DEBUG
    if (total == 4 && strcmp(self->path, "default.udbData")) 
    { 
	printf("%s at %i writing int %i\n", self->path, self->pos, *((int *)(buf))); 
    }
    #endif
    
    fseek(self->log, self->logPos, SEEK_SET);
    fwrite(&(self->pos), sizeof(long), 1, self->log);
    fwrite(&total, sizeof(size_t), 1, self->log);
    
    
    result = fwrite(buf, size, nobjs, self->log);
    self->pos += total;
    if (self->pos > self->maxPos) self->maxPos = self->pos;
    self->logPos = ftell(self->log);
    self->needsSync = 1;

    self->header.writeCount ++;
    return result;
}

size_t SearchEngine_fread(SearchEngine *self, void *buf, size_t size, size_t nobjs)
{
    size_t result;
    /* SearchEngine_sync(self);  will leave file pos in proper place */
    fseek(self->file, self->pos, SEEK_SET);
    /*printf("SearchEngine read at %i %i\n", self->pos, ftell(self->file));*/
    result = fread(buf, size, nobjs, self->file);
    self->pos = ftell(self->file);
    
    #ifdef DEBUG
    if (size == 4 && strcmp(self->path, "default.udbData")) 
    { 
	printf("%s at %i read int %i\n", self->path, self->pos, *((int *)(buf))); 
    }
    #endif
    
    return result;
}

int SearchEngine_fseek(SearchEngine *self, long offset, int origin)
{
    int result = fseek(self->file, offset, origin);
    self->pos = ftell(self->file);
    return result;
}

int SearchEngine_fputc(SearchEngine *self, int i)
{
    unsigned char c = i;
    SearchEngine_fwrite(self, &c, 1, 1);
    return i;
}

void SearchEngine_writeInt_(SearchEngine *self, int v)
{
    SearchEngine_fwrite(self, &v, sizeof(int), 1);
}

int SearchEngine_readInt(SearchEngine *self)
{
    int v;
    SearchEngine_fread(self, &v, sizeof(int), 1);
    return v;
}

void SearchEngine_setPosition_(SearchEngine *self, long pos) 
{
    self->pos = pos; 
}

long SearchEngine_position(SearchEngine *self) 
{ 
    return self->pos; 
}

long SearchEngine_setPositionToEnd(SearchEngine *self)
{
    self->pos = self->maxPos;
    return self->pos;
}

void SearchEngine_sync(SearchEngine *self)
{
    if (self->needsSync)
    {
	int i;
	long pos;
	size_t total;
	
	fflush(self->log);
	self->logPos = sizeof(SearchEngineLogHeader);
	fseek(self->log, self->logPos, SEEK_SET);

	#ifdef DEBUG
	if (strcmp(self->path, "default.udbData") != 0) 
	{ printf("------- %s - commiting %i writes -----\n", self->path, self->header.writeCount); }
	#endif

	for (i = 0; i < self->header.writeCount; i ++)
	{
	    fread(&pos,   sizeof(long), 1, self->log);
	    fread(&total, sizeof(size_t), 1, self->log);
	    self->buf = (unsigned char *)realloc(self->buf, total);
	    fread(self->buf, total, 1, self->log);
	    
	    #ifdef DEBUG
	    if (total == 4 && strcmp(self->path, "default.udbData")) 
	    { printf("%s at %i committing int %i\n", self->path, pos, *((int *)(self->buf))); }
	    #endif
	    
	    fseek(self->file, pos, SEEK_SET);
	    fwrite(self->buf, total, 1, self->file);
	}
	
	self->needsSync = 0;
	fflush(self->file);
	fseek(self->file, self->pos, SEEK_SET);
	fseek(self->log, self->logPos, SEEK_SET);
	
	SearchEngine_commitFinished(self);
	#ifdef DEBUG
	printf("--------------------------------------------------\n");
	#endif
    }
}

/* -- header ------------------------ */

void SearchEngine_readHeader(SearchEngine *self)
{
    rewind(self->log);
    fread((unsigned char *)(&(self->header)), sizeof(SearchEngineLogHeader), 1, self->log);
}

void SearchEngine_writeHeader(SearchEngine *self)
{
    rewind(self->log);
    fwrite((unsigned char *)(&(self->header)), sizeof(SearchEngineLogHeader), 1, self->log);
    fflush(self->log);
}

void SearchEngine_begin(SearchEngine *self)
{
    self->header.state = SearchEngine_UNCOMMITTED;
    self->header.writeCount = 0;
    SearchEngine_writeHeader(self);
    SearchEngine_sync(self);
}

void SearchEngine_commit(SearchEngine *self)
{
    self->header.state = SearchEngine_COMMITTED;
    SearchEngine_writeHeader(self);
    SearchEngine_sync(self);
}

void SearchEngine_clipLog(SearchEngine *self)
{
    fclose(self->log);
    truncate(self->logPath, sizeof(SearchEngineLogHeader));
    self->log = fopen(self->logPath, "r+b");
}

void SearchEngine_commitFinished(SearchEngine *self)
{
    /*SearchEngine_clipLog(self);*/
    self->header.state = SearchEngine_UNCOMMITTED;
    self->header.writeCount = 0;
    SearchEngine_writeHeader(self);
}

int SearchEngine_isCommitted(SearchEngine *self)
{
    SearchEngine_readHeader(self);
    return (self->header.state == SearchEngine_COMMITTED);
}

void SearchEngine_truncate_(SearchEngine *self, off_t size)
{
    long pos = self->pos;
    SearchEngine_commit(self);
    SearchEngine_close(self);
    /*printf("truncate(%s, %.0f)\n", self->path, (float)size);*/
    truncate(self->path, size);
    SearchEngine_open(self);
    fseek(self->file, pos, SEEK_SET);
}

int SearchEngine_Test(void)
{
    SearchEngine *t = SearchEngine_new();
    
    SearchEngine_setPath_(t, "testing");
    SearchEngine_open(t);
    SearchEngine_begin(t);
    
    {
	int i;
	
	for (i = 0; i < 3; i ++)
	{
	    char s[128];
	    
	    snprintf(s, 128, "foobar-%i", i);
	    printf("write: %s\n", s);
	    SearchEngine_fwrite(t, s, strlen(s), 1);
	}
    }
    
    printf("commit\n");
    SearchEngine_commit(t);
    printf("done\n");
    SearchEngine_close(t);
    SearchEngine_free(t);
    return 0;
}


